/*
  Name: Enemy.cpp
  Copyright: Polecatsoft 2005
  Game: Mercenary Soldier
  Author: Master Lenman
  Date: -
  Description: The enemy class is a child class of the Unit class
	  Each entity of it represents one enemy in game
*/

#include <allegro.h>
#include <string>
#include <cmath>
using namespace std;

#include "ElementaryFunctions.h"
#include "requests.h"
#include "Entity.h"
#include "Unit.h"
#include "Enemy.h"

const int DISTANCE_ALT = 150;

Enemy::Enemy() : Unit()
{
	turnDir = 0;
	shotsFired = 0;
	onScreen = false;
}

moveRequest *Enemy::getMoveRequest(int playerPosX, int playerPosY)
{
	moveRequest *mReq = NULL;
	int playerDistance;
	bool needsOnScreen = triggerDist < 0, deadPlayer = playerPosX < 0 && playerPosY < 0;
	if(!deadPlayer && (!needsOnScreen || onScreen))
		playerDistance = int(sqrt((float(playerPosX) - posX) * (float(playerPosX) - posX) + (float(playerPosY) - posY) * (float(playerPosY) - posY)));
	else
	    playerDistance = triggerDist;
	if((playerDistance < triggerDist || (onScreen && needsOnScreen)) && !deadPlayer)
	{
		AIstate = STATE_ATTACK;
		destDir = (rad2deg(atan2(playerPosX - int(posX), playerPosY - int(posY))) * -1 + 128) & 0xFF;
		if(destDir < 128)
		{
			if(direction < destDir + 128 && direction > destDir)
				turnDir = -1;
			else
				turnDir = 1;
		}
		else
		{
			if(direction < destDir && direction > destDir - 128)
				turnDir = 1;
			else
				turnDir = -1;
		}
		int i;
		for(i = 0; i < 5 && direction != destDir; ++i)
			direction = (direction + turnDir) & 0xFF;
		bool thirdParam = curWeapon == 3 || !onScreen;
		if(playerDistance > realRange[curWeapon] - DISTANCE_ALT || playerDistance < DISTANCE_ALT || thirdParam)
		{
			float newX = sin(deg2rad(direction)) * speed;
			float newY = cos(deg2rad(direction)) * speed * -1.0;
			if(playerDistance < DISTANCE_ALT)
			{
				newX *= -0.75;
				newY *= -0.75;
			}
			mReq = new moveRequest;
			mReq->set(posX, posY, posX + newX, posY + newY, 0, RIGHT_LOW, radius);

			curState = STATE_WALK;
			++phase &= 15;
		}
		else
			curState = STATE_STAND;
		if(playerDistance > realRange[curWeapon])
			AIstate = STATE_STANDBY;
	}
	else
	{
		curState = STATE_STAND;
		AIstate = STATE_STANDBY;
	}
	return mReq;
}

bulletRequest *Enemy::getBulletRequest()
{
	if(shootCooldown)
		--shootCooldown;
	bulletRequest *bReq = NULL;
	if(AIstate == STATE_ATTACK && (onScreen || curWeapon == 3))
	{
		if(!shootCooldown && destDir + 4 > direction && destDir - 4 < direction)
		{
			bReq = new bulletRequest;
			bReq->set(posX, posY, direction, curWeapon, this);
			shootCooldown += RATE[curWeapon];
			if(curState == STATE_WALK)
				curState = STATE_WALKSHOOT;
			else if(curState == STATE_STAND)
				curState = STATE_SHOOT;

			if(shootFreq)
			{
				++shotsFired;
				if(shotsFired >= shootFreq)
				{
					shotsFired = 0;
					shootCooldown += RATE[curWeapon] * 4;
				}
			}
		}
		else if(shootCooldown < 8 || RATE[curWeapon] - 8 < shootCooldown)
		{
			if(curState == STATE_STAND)
				curState = STATE_WEAPON;
		}
	}
	return bReq;
}

void Enemy::draw(int cameraX, int cameraY)
{
	Unit::draw(cameraX, cameraY);
	if(int(posX) - cameraX > -(sprites[curState]->w / 2) && int(posX) - cameraX < SCREEN_W + sprites[curState]->w / 2 && int(posY) - cameraY > -(sprites[curState]->h / 2) && int(posY) - cameraY < SCREEN_H + sprites[curState]->h / 2)
		onScreen = true;
	else
		onScreen = false;
}
