/*
  Name: SPmapClass.cpp
  Copyright: Polecatsoft 2005
  Game: Mercenary Soldier
  Author: Master Lenman
  Date: -
  Description: This is the map class for SP games, it's derived from
    the mapClass class
*/

#include <allegro.h>
#include <almp3.h>
#include <libnet.h>
#include <string>
#include <fstream>
using namespace std;

#include "ElementaryFunctions.h"
#include "requests.h"
#include "Entity.h"
#include "Unit.h"
#include "bytecode.h"
#include "Compiler.h"
#include "VM.h"
#include "Global.h"
#include "SPmapClass.h"
#include "Errors.h"

void SPmapClass::loadMap(ifstream &readfile, string filename, char gameType, VirtualMachine &VM)
{
	//Read header
	char buffer[3];
	readfile.read(buffer, 3);
	if(buffer[0] != MAJOR_VERSION || buffer[1] != MINOR_VERSION || buffer[2] != VERSION_LETTER)
		errors::mapLoadError(filename, "Version mismatch");
	readfile.read(&buffer[0], 1);
	if(buffer[0] != gameType)
		errors::mapLoadError(filename, "Wrong map type");
	mapType = buffer[0];

	//Read tileset
	string strBuffer;
	getline(readfile, strBuffer, DELIMITER);
	setTileset(strBuffer);
	animCountdown = 16;

	//Read size
	readfile.read(i2ch(width), sizeof(width));
	readfile.read(i2ch(height), sizeof(height));
	size = width * height;
	graphicData = new char[size];

	//Initialise VM for graphic command
	VM.mapSetup(width, height, graphicData);

	//Create holders for bytecode
	onEnter = new bytecode*[size];
	onLeave = new bytecode*[size];
	onUse = new bytecode*[size];
	duringTouch = new bytecode*[size];
	onExplode = new bytecode*[size];

	//string lastFive[5];

	//Load map data and scripts
	for(int i = 0; i < size; ++i)
	{
		readfile.read(&graphicData[i], 1);
		//Load script
		readfile.read(buffer, 1);

		//The file the script will be read from
		ifstream *scriptfile = &readfile;
		//The file that is created to store compiled script information
		ifstream comScript;
		//Compile script and pass it to
		if(buffer[0] == 2)
		{
			buffer[0] = 1;
			string scriptPath;
			getline(readfile, scriptPath, DELIMITER);

			/*int j;
			for(j = 0; j < 5; ++j)
			{
				if(lastFive[j] == scriptPath)
				{
					onEnter[i] = onEnter[i - j - 1];
					onUse[i] = onUse[i - j - 1];
					onLeave[i] = onLeave[i - j - 1];
					duringTouch[i] = duringTouch[i - j - 1];
					onExplode[i] = onExplode[i - j - 1];
					break;
				}
			}
			for(j = 4; j > 0; --j)
				lastFive[j] = lastFive[j - 1];
			lastFive[0] = scriptPath;*/

			if(!compile(scriptPath, "scripts/temp.mss"))
				errors::scriptError("An error occured in compile stage\nSee compiler.log for details.");

			comScript.open("scripts/temp.mss", ios::binary);
			if(comScript.fail())
				errors::fatalError("Fatal error: Opening 'scripts/temp.mss' failed!");

			scriptfile = &comScript;
		}
		if(buffer[0] == 1)
		{
			unsigned char fullCode[0x400]; //1KB buffer for script data
			int pos = 0;
			while(true)
			{
				scriptfile->read((char *)(&fullCode[pos]), 1);
				if(fullCode[pos] == SCRIPTEND)
					break;
				++pos;
				if(pos >= 1024)
					errors::scriptError("Script too big");
			}
			if(scriptfile != &readfile)
				scriptfile->close();
			//Buffer for code
			bytecode codeBuffer;
			//onLoad
			codeBuffer.extract(fullCode, ONLOAD);
			if(codeBuffer.getLength())
			{
				//The standard now defines that the variables 1-3 are only availible ingame
				//int x, y;
				//xyFromOffset(x, y, pos);
				//VM.getVar(1) = x;
				//VM.getVar(2) = y;
				//VM.getVar(3) = graphicData[pos];
				VM.compute(codeBuffer);
			}
			//onEnter
			codeBuffer.extract(fullCode, ONENTER);
			if(codeBuffer.getLength())
			{
				onEnter[i] = new bytecode;
				//Copy content of codeBuffer
				*onEnter[i] = codeBuffer;
			}
			else
				onEnter[i] = NULL;
			//onUse
			codeBuffer.extract(fullCode, ONUSE);
			if(codeBuffer.getLength())
			{
				onUse[i] = new bytecode;
				*onUse[i] = codeBuffer;
			}
			else
				onUse[i] = NULL;
			//onExplode
			codeBuffer.extract(fullCode, ONEXPLODE);
			if(codeBuffer.getLength())
			{
				onExplode[i] = new bytecode;
				*onExplode[i] = codeBuffer;
			}
			else
				onExplode[i] = NULL;
			//duringTouch
			codeBuffer.extract(fullCode, DURINGTOUCH);
			if(codeBuffer.getLength())
			{
				duringTouch[i] = new bytecode;
				*duringTouch[i] = codeBuffer;
			}
			else
				duringTouch[i] = NULL;
			//onLeave
			codeBuffer.extract(fullCode, ONLEAVE);
			if(codeBuffer.getLength())
			{
				onLeave[i] = new bytecode;
				*onLeave[i] = codeBuffer;
			}
			else
				onLeave[i] = NULL;
		}
		else
		{
			onEnter[i] = NULL;
			onUse[i] = NULL;
			onLeave[i] = NULL;
			duringTouch[i] = NULL;
			onExplode[i] = NULL;
		}
	}
	visibleTilesX = SCREEN_W / TILESIZE + 1;
	if(visibleTilesX > width)
		visibleTilesX = width;
	visibleTilesY = SCREEN_H / TILESIZE + 1;
	if(visibleTilesY > height)
		visibleTilesY = height;
	maxX = width * TILESIZE - SCREEN_W / 2;
	maxY = height * TILESIZE - SCREEN_H / 2;
}

BITMAP *SPmapClass::getMapPreview(int sizeX, int sizeY)
{
	if(preview)
		return preview;
	BITMAP *temp = tileset;

	smallSize = sizeX / width;
	if(sizeY / height < smallSize)
		smallSize = sizeY / height;
	tileset = createBitmap(smallSize * tileTypes, smallSize);
	stretch_blit(temp, tileset, 0, 0, temp->w, temp->h, 0, 0, tileset->w, tileset->h);

	preview = createBitmap(smallSize * width, smallSize * height);

	//int playerTileX = playerPosX >> 5, playerTileY = playerPosY >> 5;

	for(int i = 0; i < width; ++i)
	{
		for(int j = 0; j < height; ++j)
		{
			drawTile(i * smallSize, j * smallSize, graphicData[getOffset(i, j)], smallSize, preview);
		}
	}

	destroy_bitmap(tileset);
	tileset = temp;
	return preview;
}

int SPmapClass::getSmallSize()
{
	return smallSize;
}

SPmapClass::~SPmapClass()
{
	if(onEnter)
		delete [] onEnter;
	if(onLeave)
		delete [] onLeave;
	if(onUse)
		delete [] onUse;
	if(duringTouch)
		delete [] duringTouch;
	if(onExplode)
		delete [] onExplode;
	if(preview)
		destroy_bitmap(preview);
}

void SPmapClass::handleMoveRequest(moveRequest *mReq, VirtualMachine &VM, int onUseScript)
{
	if(!mReq)
		return;

	int oldTileX = int(mReq->getOldPosX()) >> 5,
		oldTileY = int(mReq->getOldPosY()) >> 5;

	//Exception for player requests when not moving
	if(!mReq->wasRequested())
	{
		VM.getVar(1) = oldTileX;
		VM.getVar(2) = oldTileY;
		VM.getVar(3) = graphicData[getOffset(oldTileX, oldTileY)];
		if(duringTouch[getOffset(oldTileX, oldTileY)])
			VM.compute(*duringTouch[getOffset(oldTileX, oldTileY)]);
		if(mReq->getScripts() & SCRIPT_USE && onUse[getOffset(oldTileX, oldTileY)])
			VM.compute(*onUse[getOffset(oldTileX, oldTileY)]);
		return;
	}

	int newTileX = int(mReq->getNewPosX()) >> 5,
		newTileY = int(mReq->getNewPosY()) >> 5;

	//Decline the request if the object wants to leave the map
	if(mReq->getNewPosX() - mReq->getRadius() < 0.0 || mReq->getNewPosX() + mReq->getRadius() > float(width * TILESIZE))
	{
		mReq->declineX();
	}
	if(mReq->getNewPosY() - mReq->getRadius() < 0.0 || mReq->getNewPosY() + mReq->getRadius() > float(height * TILESIZE))
	{
		mReq->declineY();
	}


	//Decline the request if the object's rights are equal or lower to the new tile's rights
	if(mReq->getRights() <= hightData[graphicData[getOffset(newTileX, newTileY)]])
	{
		mReq->decline();
	}

	//Execute scripts
	if(mReq->getScripts() & SCRIPT_PLAYER)
	{
		if(!mReq->wasDeclined())
		{
			VM.getVar(1) = newTileX;
			VM.getVar(2) = newTileY;
			VM.getVar(3) = graphicData[getOffset(newTileX, newTileY)];

			if(duringTouch[getOffset(newTileX, newTileY)])
				VM.compute(*duringTouch[getOffset(newTileX, newTileY)]);
			if(onUse[getOffset(newTileX, newTileY)] && onUseScript)
				VM.compute(*onUse[getOffset(newTileX, newTileY)]);
			if(mReq->getScripts() & SCRIPT_USE && onUse[getOffset(newTileX, newTileY)])
				VM.compute(*onUse[getOffset(newTileX, newTileY)]);
			if(newTileX != oldTileX || newTileY != oldTileY)
			{
				if(onEnter[getOffset(newTileX, newTileY)])
					VM.compute(*onEnter[getOffset(newTileX, newTileY)]);

				VM.getVar(1) = oldTileX;
				VM.getVar(2) = oldTileY;
				VM.getVar(3) = graphicData[getOffset(oldTileX, oldTileY)];

				if(onLeave[getOffset(oldTileX, oldTileY)])
					VM.compute(*onLeave[getOffset(oldTileX, oldTileY)]);
			}
		}
		else
		{
			VM.getVar(1) = oldTileX;
			VM.getVar(2) = oldTileY;
			VM.getVar(3) = graphicData[getOffset(oldTileX, oldTileY)];

			if(onUse[getOffset(oldTileX, oldTileY)] && onUseScript)
				VM.compute(*onUse[getOffset(oldTileX, oldTileY)]);
			if(duringTouch[getOffset(oldTileX, oldTileY)])
				VM.compute(*duringTouch[getOffset(oldTileX, oldTileY)]);
			if(mReq->getScripts() & SCRIPT_USE && onUse[getOffset(oldTileX, oldTileY)])
				VM.compute(*onUse[getOffset(oldTileX, oldTileY)]);
		}
	}
	if(mReq->wasDeclined() && mReq->getScripts() & SCRIPT_EXPLODE && onExplode[getOffset(newTileX, newTileY)])
	{
		VM.getVar(1) = newTileX;
		VM.getVar(2) = newTileY;
		VM.getVar(3) = graphicData[getOffset(newTileX, newTileY)];
		VM.compute(*onExplode[getOffset(newTileX, newTileY)]);
	}
}
