/*
  Name: bytecode.cpp
  Copyright: Polecatsoft 2005
  Game: Mercenary Soldier
  Author: Master Lenman
  Date: -
  Description: This class contains byte code for a single event
    It can be fed into the virtual machine
*/

#include <allegro.h>
#include <string>
#include <fstream>
using namespace std;

#include "Compiler.h"
#include "bytecode.h"
#include "Errors.h"

bytecode::bytecode()
{
	data = NULL;
	length = 0;
	pos = 0;
}

bytecode::bytecode(unsigned char *fullCode, int event)
{
	data = NULL;
	length = 0;
	pos = 0;
	extract(fullCode, event);
}

void bytecode::extract(unsigned char *fullCode, int event)
{
	//char buffer;
	int rPos = 0;
	int scriptPos = -1;
	int scriptSize = 0;
	if(data)
		delete [] data;
	data = NULL;
	length = 0;
	pos = 0;
	while(true)
	{
		if(fullCode[rPos] == SCRIPTEND)
			break;
		if(fullCode[rPos] == EVENT)
		{
			++rPos;
			if(fullCode[rPos] == event)
			{
				++rPos;
				scriptPos = rPos;
				while(true)
				{
					if(fullCode[rPos] == EVENTEND)
					{
						++scriptSize;
						break;
					}
					++rPos;
					++scriptSize;
				}
				break;
			}
			else
			{
				while(fullCode[rPos] != EVENTEND)
				{
					++rPos;
				}
				++rPos;
			}
		}
	}
	if(scriptPos >= 0)
	{
		data = new unsigned char[scriptSize];
		length = scriptSize;
		int i;
		for(i = 0; i < scriptSize - 1; ++i)
		{
			data[i] = fullCode[scriptPos + i];
		}
		data[i] = SCRIPTEND;
	}
	if(event == 9)
	{
		ofstream output("globalScript.mss", ios::binary);
		output.write((char*)data, length);
		output.close();
	}
}

bytecode::~bytecode()
{
	if(data)
		delete [] data;
}

int bytecode::getLength()
{
	return length;
}

void bytecode::reset()
{
	pos = 0;
}

void bytecode::read(unsigned char *dest, int bytes)
{
	if(pos >= length)
	{
		int breakpoint = 0;
		//errors::scriptError("Bytecode overflow!");
	}
	for(int i = 0; i < bytes; ++i)
		dest[i] = data[pos + i];
	pos += bytes;
}

void bytecode::read(char *dest, int bytes)
{
	read((unsigned char*)dest, bytes);
}

void bytecode::operator =(bytecode &code)
{
	if(data)
		delete [] data;
	data = new unsigned char[code.getLength()];
	length = code.getLength();
	code.reset();
	code.read(data, code.getLength());
}

int bytecode::getPos()
{
	return pos;
}

bool bytecode::isVar()
{
	char temp;
	read(&temp, 1);
	if(temp == ISVAR)
		return true;
	else if(temp == ISNUM)
		return false;
	else
		errors::scriptError("Unexpected value for argument identifier!");
	return false;
}

int bytecode::getConstPar()
{
	int temp;
	read((char*)&temp, sizeof(int));
	return temp;
}

int bytecode::getVarPar()
{
	unsigned char temp;
	read((char*)&temp, 1);
	return int(temp);
}

string bytecode::getString(char limiter)
{
	string temp;
	char chBuf = 1;
	while(true)
	{
		read(&chBuf, 1);
		if(chBuf == limiter)
			break;
		temp.append(1, chBuf);
	}
	return temp;
}
